
import sys
import cv2
import random
import numpy as np
from sout import sout
# 共通部品 [utils.py]
from .parts.utils import show_img, TileArt
# デバッグ用のプレーヤー [debug_players.py]
from .parts.debug_players import RandomPlayer, HumanPlayer
# Deep Q Network AI [DQN_Agent]
from .parts.DQN_Agent import DQN_Agent
# Catcherゲーム [catcher_game]
from .parts import catcher_game
# # 紅白ゲーム [rw_game]
# from .parts import rw_game




import sys
# from ezRL import catcher_game as c_game	# 本番環境での使用法
from relpath import add_import_path
add_import_path("../")
# Catcherゲーム [catcher_game]
from _develop_ezRL import catcher_game as c_game
# Deep Q Network AI [DQN_Agent]
from _develop_ezRL import DQN_Agent

# debug
from _develop_ezRL import TileArt
# タイルを並べた画像の作成 [ezRL]
t_art = TileArt(
	canvas_size = (300,300),
	tile_file_dic = {
		"wall": "./test_tiles/wall.bmp",
		"box": "./test_tiles/box.bmp",
		"player": "./test_tiles/player.png"
	},
	transparent_color = (255, 0, 255)	# 透過色の指定 (b,g,r)
)
# タイルの表示サイズ
tile_size = 64
# タイルを並べた画像の作成
img = t_art.gen_img([
	{"pos": (0,0), "size": (tile_size, tile_size), "texture": "wall", "layer": 1},
	{"pos": (32,0), "size": (tile_size, tile_size), "texture": "player", "layer": 2},
	{"pos": (64,64), "size": (tile_size, tile_size), "texture": "box", "layer": 1},
])
# 表示
from _develop_ezRL import show_img
show_img(img)
sys.exit()

train_ai = DQN_Agent(action_ls = c_game.action_ls, ai_obs = c_game.ai_obs) # Deep Q Network AI
episode_n = 700	# エピソード数
total_reward_ls = []
for episode_idx in range(episode_n):
	# state, actionの初期化
	action, state = "initial_action", c_game.gen_init_state(game_params = {})
	reward_ls = []
	# ゲーム進行
	while state["finished"] is False:
		state, reward = c_game.game_step(state, action)
		action = train_ai.think(state, reward) # 行動決定
		reward_ls.append(reward)
	total_reward_ls.append(sum(reward_ls))
	print("Episode #%d, Reward: %.1f"%(episode_idx, total_reward_ls[-1]))
# 獲得報酬の推移を表示
from matplotlib import pyplot as plt
plt.plot(total_reward_ls)
plt.show()
# state, actionの初期化
action, state = "initial_action", c_game.gen_init_state(game_params = {})
# テストプレイ
while state["finished"] is False:
	human_obs(state, waitKey_msec = 100) # 人間用の観測関数
	state, reward = c_game.game_step(state, action)
	action = train_ai.think(state, reward)


sys.exit()

#### 以下、ちゃんとtestとtrainが分かれている例


from __init__ import RandomPlayer	# 開発時




# ランダムAI
train_ai = RandomPlayer(action_ls = ["act_0", "act_1"])

# 学習
for _ in range(3):
	action = train_ai.think(state = None, reward = 1)
	print(action)

# テスト用プレーヤーを生成
test_ai = train_ai.gen_test()

# テスト
for _ in range(3):
	action = test_ai.think(state = None, reward = 1)
	print(action)

sys.exit()

######### 以下、未整備

# # debug用人間プレーヤー
# from ezRL import HumanPlayer
# # 学習用AI, 対戦フィールド (arena)
# from ezRL import DQN_AI, arena
# # テスト用ゲーム (紅白ゲーム)
# from ezRL import RW_Game

# # # debug用人間プレーヤー
# # human_p = HumanPlayer(action_dic = {"Red": [1,0], "White": [0,1]})
# # game = RW_Game(round_n = 5)	# テスト用ゲーム (紅白ゲーム)
# # # 対戦して学習
# # arena(game, [human_p])	# 対戦フィールド (arena)
# # sys.exit()

# train_ai = DQN_AI(action_dim = 2)	# 学習用AI
# game = RW_Game(round_n = 100)	# テスト用ゲーム (紅白ゲーム)
# # 対戦して学習
# arena(game, [train_ai])	# 対戦フィールド (arena)

# # テスト用プレーヤーを生成
# test_ai = train_ai.gen_test()
# game = RW_Game(round_n = 100)	# テスト用ゲーム (紅白ゲーム)
# # 対戦して性能テスト (学習済みAI)
# arena(game, [test_ai])	# 対戦フィールド (arena)




import sys
from .utils import show_img 

# ランダムAI
class RandomPlayer:
	# 初期化処理
	def __init__(self, action_ls):
		self.action_ls = action_ls
	# テスト用プレーヤーを生成
	def gen_test(self):
		test_ai = self
		return test_ai
	# アクションを考える
	def think(self, state, reward):
		action = random.choice(self.action_ls)
		return action

# 人間プレーヤー [ezRL]
class HumanPlayer:
	# 初期化処理
	def __init__(self, action_ls, human_obs):
		self.action_ls = action_ls
		self.human_obs = human_obs
	# アクションを考える
	def think(self, state, reward):
		print("報酬を%.2fもらえました"%reward)
		obs_img = self.human_obs(state)
		show_img(obs_img, waitKey = 100, ratio = 0.5, interpolation = "INTER_CUBIC")	# 状態の表示
		while True:
			print("action: %s"%(", ".join(self.action_ls)))
			action = input("action>").strip()
			if action in self.action_ls: return action
			print("【！】actionが不正です")




# Deep Q Network AI [DQN_Agent]

import sys
import numpy as np
import random
from collections import deque
from relpath import add_import_path
add_import_path("../")
# シンプルなCNN [normal_cnn]
from normal_cnn import Normal_CNN

# 攪拌 + リサンプリングするためのインデックスのリストを生成
def gen_random_idx_ls(n, ratio):
	idx_ls = random.sample(list(range(n)), k = int(n * ratio))
	return idx_ls

# Deep Q Network AI [DQN_Agent]
class DQN_Agent:
	# 初期化処理
	def __init__(self, action_ls, ai_obs, 
				learning_interval_epi = 3, clone_interval_epi = 10,
				gamma = 0.9, buffer_size = 5000,
				eps_start = 0.5, eps_end = 0.0, eps_decay_end_epi_n = 500):
		# ゲーム側から提供される情報
		self.action_ls = action_ls	# action一覧
		self.action_idx_dict = {a:i for i, a in enumerate(action_ls)} # アクション名から通し番号を取り出す辞書
		self.ai_obs = ai_obs	# AI用の観測関数
		# 各種パラメータ
		self.learning_interval_epi = learning_interval_epi	# 学習をするまでのエピソード数
		self.clone_interval_epi = clone_interval_epi	# NNを入れ替えるまでのエピソード数
		self.eps_start = eps_start	# 開始時のランダムなactionをとる確率
		self.eps_end = eps_end	# 終了時のランダムなactionをとる確率
		self.eps_decay_end_epi_n = eps_decay_end_epi_n	# Epsilonが安定するまでのエピソード数
		self.eps = self.eps_start	# ランダムなactionをとる確率
		self.gamma = gamma	# 割引率
		self.train_data_ratio = 0.1 # Experience Replay時にlogから取り出すデータの割合
		# 学習に使うデータの格納場所
		self.playlog_ls = deque([{"now_finished": True}], maxlen = buffer_size)	# プレイ記録の格納場所
		# 内部状態
		self.nn_initialized = False	# NNが初期化されているか
		self.episode_cnt = 0	# 累積エピソード回数
		# ニューラルネットの構成
		nn_param = {
			"conv_layer_info": [
				# {"n_filter": 8, "kernel_size": (5, 5)},
				# {"n_filter": 16, "kernel_size": (5, 5)},
				# {"n_filter": 32, "kernel_size": (5, 5)},
			],
			"dense_layer_info": [
				{"unit_n": 512, "activation": "relu"},
				{"unit_n": 128, "activation": "relu"},
				{"unit_n": 32, "activation": "relu"},
				{"unit_n": 8, "activation": "relu"},
			],
			"output_activation": None
		}
		# Q関数を近似するニューラルネットの定義
		from keras.optimizers import RMSprop
		self.q_network = Normal_CNN(len(action_ls),
									nn_param,
									loss = "mse",
									optimizer = RMSprop())	# シンプルなCNN [normal_cnn]
	# 行動方策
	def policy(self, obs_img):
		# Epsilon Greedyにより、一定確率でランダムな行動を選ぶ
		if random.random() < self.eps:
			# ランダムに行動を選ぶ
			selected_a = random.choice(self.action_ls)
		else:
			# Q値が最も高いactionを選ぶ
			q_ls = self.q_network.predict(np.array([obs_img]))[0]
			best_a_idx = max([(i, q) for i, q in enumerate(q_ls)], key = lambda x: x[1])[0]
			selected_a = self.action_ls[best_a_idx]
		return selected_a
	# 行動選択、記憶、学習
	def think(self, state, reward):
		# 観測
		obs_img = self.ai_obs(state)
		# q_networkが未初期化の場合、初期化
		if self.nn_initialized is False:
			self.q_network.train_on_batch(np.array([obs_img]), np.array([[0 for _ in self.action_ls]]))
			from keras.models import clone_model
			self.target_network = clone_model(self.q_network.keras_model)
			self.nn_initialized = True
		# 行動方策による行動の選択
		selected_a = self.policy(obs_img)
		# 一つ前Stepのログの要素を参照する
		def get_prev_log(key):
			if self.playlog_ls[-1]["now_finished"] is True: return None
			return self.playlog_ls[-1][key]
		if state["finished"] is True: selected_a = "finish_action"
		# 結果の追記
		self.playlog_ls.append({
			"prev_obs": get_prev_log("now_obs"), "prev_action": get_prev_log("now_action"),
			"now_obs": obs_img,	"now_action": selected_a,
			"now_reward": reward, "now_finished": state["finished"]
		})
		# エピソード終了に伴う処理
		if state["finished"] is True: self.terminate_episode()
		return selected_a	# 選択したactionを返す
	# エピソード終了に伴う処理
	def terminate_episode(self):
		# エピソード数のインクリメント
		self.episode_cnt += 1
		# モデルの学習
		if self.episode_cnt % self.learning_interval_epi == 0:
			self.train_model()
		# ターゲットモデルの更新
		if self.episode_cnt % self.clone_interval_epi == 0:
			self.update_target_network()
		# Epsilonの更新
		self.update_eps()
	# モデルの学習
	def train_model(self):
		filtered_playlog_ls = [e for e in self.playlog_ls if e.get("prev_obs", None) is not None]
		# 学習データ(入力側)の生成
		input_ls = np.array([one_log["prev_obs"] for one_log in filtered_playlog_ls])
		# 学習データ(出力側)の生成
		esitimated_ls = self.target_network.predict(input_ls)	# 学習の安定のため、重みを固定したNNによってQ値を計算
		output_ls = np.array(esitimated_ls)
		for log_idx, one_log in enumerate(filtered_playlog_ls):
			# エピソード終了の1つ前のstateのみ遷移先基準のQ値が計算できないため、条件分岐
			if one_log["now_finished"] is True:
				gain = one_log["now_reward"]
			else:
				next_s_max_q = max(esitimated_ls[log_idx + 1])
				gain = one_log["now_reward"] + self.gamma * next_s_max_q
			# 実際に取った行動のQ値のみ更新
			action_idx = self.action_idx_dict[one_log["prev_action"]]	# アクション名から通し番号を取り出す辞書
			output_ls[log_idx][action_idx] = gain
		# 攪拌 + リサンプリングして学習
		idx_ls = gen_random_idx_ls(n = len(input_ls), ratio = self.train_data_ratio)	# 攪拌 + リサンプリングするためのインデックスのリストを生成
		loss = self.q_network.train_on_batch(input_ls[idx_ls], output_ls[idx_ls])
		# print("loss: %.4f"%loss)
	# ターゲットモデルの更新
	def update_target_network(self):
		self.target_network.set_weights(self.q_network.keras_model.get_weights())
		# print("teacher model is updated.")
	# Epsilonの更新
	def update_eps(self):
		eps_s, eps_e = self.eps_start, self.eps_end
		eps_cnt = self.episode_cnt
		decay_end_epi_n = self.eps_decay_end_epi_n
		if self.episode_cnt >= decay_end_epi_n:
			self.eps = eps_e
		else:
			self.eps = eps_s - eps_cnt * (eps_s - eps_e) / decay_end_epi_n
		# print("Epsilon: %.2f"%self.eps)



# Catcherゲーム [catcher_game]

import sys
import copy
import random
import numpy as np

# 全てのアクション [catcher_game]
action_ls = ["Right", "NoMove", "Left"]

# 初期状態の生成 [catcher_game]
def gen_init_state(game_params):
	# 画面のサイズの定義
	w = 7
	h = 7
	board_size = {"w": w, "h": h}
	# パドルの定義
	paddle_size = 3	# パドルサイズ
	paddle_pos = int(random.random() * (w - paddle_size))	# パドル位置
	# フルーツの数・位置に関する定義
	fruits_n = 20
	fruits_delta = h // 3	# フルーツ同士の間隔
	head_margin, tail_margin = h // 2, h # テープ両端のフルーツがない部分の長さ
	# フルーツ位置を保持するテープの生成
	random_pos = lambda: random.choice(list(range(w)))	# フルーツの場所のランダム生成
	none_ls = lambda n: [None for _ in range(n)]	# Noneのリストを生成 (要素数n)
	tape_info = []
	tape_info += none_ls(n = head_margin)
	for fruits_idx in range(fruits_n - 1):
		tape_info.append(random_pos())
		tape_info += none_ls(n = fruits_delta - 1)
	tape_info.append(random_pos())
	tape_info += none_ls(n = tail_margin)
	# 初期state
	init_state = {
		"finished": False,
		"cursor": 0,
		"board_size": board_size,
		"paddle_size": paddle_size,
		"paddle_pos": paddle_pos,
		"tape_info": tape_info,
	}
	return init_state

# ゲーム進行関数 [catcher_game]
def game_step(state, action):
	# ゲーム開始時の空打ちのaction
	if action == "initial_action":
		return state, 0
	# stateからの情報の取り出し
	paddle_pos = state["paddle_pos"]
	paddle_size = state["paddle_size"]
	w = state["board_size"]["w"]
	cursor = state["cursor"]
	tape_info = state["tape_info"]
	# 通常のゲーム進行
	if action == "Left":
		next_paddle_pos = max(0, paddle_pos - 1)
	elif action == "Right":
		next_paddle_pos = min(w - paddle_size, paddle_pos + 1)
	elif action == "NoMove":
		next_paddle_pos = paddle_pos
	else:
		raise Exception("[error] 不正なaction(%s)が指定されました"%action)
	next_cursor = cursor + 1
	# 次のstateの作成
	next_state = {k:state[k] for k in state}	# Shallow copy
	next_state["cursor"] = next_cursor
	next_state["paddle_pos"] = next_paddle_pos
	# ゲーム終了条件
	if next_cursor == len(tape_info):
		next_state["finished"] = True
		return next_state, 0
	# フルーツ獲得の判定
	reward = 0
	roi_tape = tape_info[next_cursor]	# 注視領域
	if roi_tape is not None:
		# パドルの領域にフルーツが入っていた場合
		if (next_paddle_pos <= roi_tape) & (roi_tape < next_paddle_pos + paddle_size):
			reward = 1
		else:
			reward = -1
	return next_state, reward

# AI用の観測関数 [catcher_game]
def ai_obs(state):
	# stateからの情報の取り出し
	w, h = state["board_size"]["w"], state["board_size"]["h"]
	paddle_pos = state["paddle_pos"]
	paddle_size = state["paddle_size"]
	tape_info = state["tape_info"]
	cursor = state["cursor"]
	# 返す画像の初期化
	ret_img = np.zeros((h, w, 3), dtype = "float64")
	# フルーツの描画
	partial_tape = tape_info[cursor:cursor + h]
	fruit_info = [(i, p) for i, p in enumerate(partial_tape) if p is not None]
	for one_info in fruit_info:
		i, p = one_info
		ret_img[- i - 1, p, :] = [0.5, 0.25, 1.0]
	# パドルの描画
	ret_img[- 1, paddle_pos:paddle_pos + paddle_size, :] = [1.0, 1.0, 1.0]
	return ret_img

# float64型からuint8型への変換
def float2uint8(float_img):
	clipped_img = copy.deepcopy(float_img)
	clipped_img[clipped_img < 0] = 0
	clipped_img[clipped_img > 1] = 1
	uint_img = np.array(clipped_img * 255, dtype = "uint8")
	return uint_img

# 人間用の観測関数 [catcher_game]
def human_obs(state, waitKey_msec = 1):
	import cv2
	float_img = ai_obs(state)	# AI用の観測関数 [catcher_game]
	uint_img = float2uint8(float_img)	# float64型からuint8型への変換
	resized = cv2.resize(uint_img, None, fx = 30, fy = 30, interpolation = False)
	cv2.imshow("human_obs", resized)
	cv2.waitKey(waitKey_msec)



# 紅白ゲーム [rw_game]

raise Exception("mijisso!")

########## 以下、古い実装

# # テスト用ゲーム (紅白ゲーム)
# class RW_Game:
# 	# 初期化処理
# 	def __init__(self, round_n):
# 		self.round_n = round_n
# 		self.q_idx = 0	# 何問目か
# 		# 次の正解を作る
# 		self.renew_ans()
# 	# 現在のプレーヤーを返す
# 	def now_player(self):
# 		return 0	# 1人用ゲームなので手番は変化しない
# 	# 観測
# 	def observe(self, p_id = None):
# 		# 終了している場合はobserveできない
# 		if self.finished() is True: raise Exception("[error] 既にゲームが終了しています。")
# 		# p_id指定なし(現在の手番)の場合は、p_id==0とする(1人用ゲームなので)
# 		if p_id is None: p_id = 0
# 		# 画像を作る
# 		obs_img = np.zeros((200,200,3), dtype = np.uint8)
# 		obs_img[:,:,:] = 220	# 全て白で塗る
# 		if self.true_ans == 0:
# 			obs_img[:,:,:-1] = 30
# 		return obs_img
# 	# アクションの実行
# 	def action(self, action):
# 		# 終了している場合はactionできない
# 		if self.finished() is True: raise Exception("[error] 既にゲームが終了しています。")
# 		# actionの型の判定
# 		if len(action) != 2: raise Exception("[error] actionの型が不正です")
# 		# action配列のデコード
# 		r_value, w_value = action
# 		rw_action = int(w_value > r_value)
# 		# rewardの設定
# 		reward = int(self.true_ans == rw_action)
# 		# 出題数をインクリメント
# 		self.q_idx += 1
# 		# 次の正解を作る
# 		self.renew_ans()
# 		return reward
# 	# 次の正解を作る
# 	def renew_ans(self):
# 		self.true_ans = int(random.random()*2)
# 	# 終了判定
# 	def finished(self):
# 		if self.q_idx >= self.round_n: return True
# 		return False
